from typing import Iterable, List

import numpy as np
import onnx

from furiosa.optimizer.frontend.onnx.transformer import ONNXTransformer, utils
from furiosa.optimizer.interfaces.transformer import Transformer


class FuseGatherMatMul(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:  # pylint: disable=no-member
        for transformer in [
            Pattern_1,
        ]:
            model = transformer(model).transform()
        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Gather --> MatMul --> next
    to
        prev --> Gather --> next

    if 1. MatMul must have exactly one initializer
       2. Gather.data must be defined in graph.initializer
       3. MatMul weight's data_type == onnx.TensorProto.FLOAT
       4. rank(MatMul weight) == 2
       5. Gather.data's data_type == onnx.TensorProto.FLOAT
       6. rank(Gather.data) == 2
       7. (Gather.axis == 0 and Matmul.input[1] is initializer) or
          (Gather.axis == 1 and Matmul.input[0] is initializer)
    """

    pattern_to_match = ['Gather', 'MatMul']

    def pattern_matching(
        self, base_node: onnx.NodeProto  # pylint: disable=no-member
    ) -> Iterable[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)

        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
        )

        gather, _ = matched_nodes
        return gather.input

    # pylint: disable=too-many-return-statements
    def pattern_condition_checker(
        self, nodes_to_check: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> bool:
        gather, matmul = nodes_to_check

        if (matmul.input[0] in self.initializer_map) == (matmul.input[1] in self.initializer_map):
            return False

        matmul_weight = self.get_init_node_input(matmul)

        gather_data = gather.input[0]
        if gather_data not in self.initializer_map:
            return False

        # pylint: disable-next=no-member
        if self.get_value_info_dtype(matmul_weight) != onnx.TensorProto.FLOAT:
            return False

        if len(self.get_value_info_shape(matmul_weight)) != 2:
            return False

        # pylint: disable-next=no-member
        if self.get_value_info_dtype(gather_data) != onnx.TensorProto.FLOAT:
            return False

        if len(self.get_value_info_shape(gather_data)) != 2:
            return False

        gather_axis = utils.get_attribute(gather.attribute, "axis", default=0)
        if not (
            (gather_axis == 0 and matmul.input[1] in self.initializer_map)
            or (gather_axis == 1 and matmul.input[0] in self.initializer_map)
        ):
            return False
        return True

    @staticmethod
    def make_new_node(
        matched_nodes: Iterable[onnx.NodeProto],  # pylint: disable=no-member
    ) -> List[onnx.NodeProto]:  # pylint: disable=no-member
        gather, matmul = matched_nodes

        return [
            onnx.helper.make_node(
                'Gather',
                inputs=[gather.input[0] + '_fused', gather.input[1]],
                outputs=[matmul.output[0]],
                name=matmul.output[0] + '_1',
                **utils.get_node_attributes(gather),
            )
        ]

    def make_new_init(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.TensorProto]:  # pylint: disable=no-member
        gather, matmul = matched_nodes

        table_tensor = gather.input[0]
        weight_tensor = self.get_init_node_input(matmul)

        table_arr = self.get_initializer_array(table_tensor)
        weight_arr = self.get_initializer_array(weight_tensor)

        init_idx = list(matmul.input).index(weight_tensor)
        if init_idx == 1:
            fused_table_arr = np.matmul(table_arr, weight_arr)
        elif init_idx == 0:
            fused_table_arr = np.matmul(weight_arr, table_arr)

        return [onnx.numpy_helper.from_array(fused_table_arr, table_tensor + '_fused')]



"""FuriosaAI optimizer"""

import furiosa.common.utils
from furiosa.optimizer.frontend.onnx import optimize_model

__version__ = furiosa.common.utils.get_sdk_version(__name__)

__all__ = ["optimize_model"]



from typing import Iterable, List

import numpy as np
import onnx

from furiosa.optimizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.optimizer.interfaces.transformer import Transformer


class ConvertPReluToRelu(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:  # pylint: disable=no-member
        for transformer in [Pattern_1, Pattern_2]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        PRelu(x) = slope * x if x < 0, x if x >=0
    into
        (1 - slope) * Relu(x) + slope * x
    if
        1. PRelu's input[1] is an initializer
    """

    pattern_to_match = ['PRelu']

    def pattern_matching(
        self, base_node: onnx.NodeProto  # pylint: disable=no-member
    ) -> Iterable[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
            vis_to_add=self.make_new_vi(matched_nodes),
        )

        return base_node.input

    def pattern_condition_checker(
        self, nodes_to_check: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> bool:
        (prelu,) = nodes_to_check

        return prelu.input[1] in self.initializer_map

    def make_new_node(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.NodeProto]:  # pylint: disable=no-member
        (prelu,) = matched_nodes

        relu = onnx.helper.make_node(
            'Relu',
            inputs=[prelu.input[0]],
            outputs=[prelu.output[0] + '_relu_out'],
            name=prelu.output[0] + '_0',
        )

        mul_0 = onnx.helper.make_node(
            'Mul',
            inputs=[prelu.output[0] + '_relu_out', prelu.output[0] + '_slope_fused'],
            outputs=[prelu.output[0] + '_mul_0_out'],
            name=prelu.output[0] + '_1',
        )

        mul_1 = onnx.helper.make_node(
            'Mul',
            inputs=[prelu.input[0], prelu.input[1]],
            outputs=[prelu.output[0] + '_mul_1_out'],
            name=prelu.output[0] + '_2',
        )

        add = onnx.helper.make_node(
            'Add',
            inputs=[prelu.output[0] + '_mul_0_out', prelu.output[0] + '_mul_1_out'],
            outputs=prelu.output,
            name=prelu.output[0] + '_3',
        )

        return [relu, mul_0, mul_1, add]

    def make_new_init(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.TensorProto]:  # pylint: disable=no-member
        (prelu,) = matched_nodes

        slope = self.get_initializer_array(prelu.input[1])
        one_minus_slope = np.ones_like(slope) - slope
        new_init = onnx.numpy_helper.from_array(one_minus_slope, prelu.output[0] + '_slope_fused')

        return [new_init]

    def make_new_vi(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.ValueInfoProto]:  # pylint: disable=no-member
        (prelu,) = matched_nodes
        input_shape = self.get_value_info_shape(prelu.input[0])
        input_dtype = self.get_value_info_dtype(prelu.input[0])

        return [
            onnx.helper.make_tensor_value_info(
                prelu.output[0] + '_relu_out',
                input_dtype,
                input_shape,
            ),
            onnx.helper.make_tensor_value_info(
                prelu.output[0] + '_mul_0_out',
                input_dtype,
                input_shape,
            ),
            onnx.helper.make_tensor_value_info(
                prelu.output[0] + '_mul_1_out',
                input_dtype,
                input_shape,
            ),
        ]


class Pattern_2(ONNXTransformer):
    """
    transform
        PRelu(x) = slope * x if x < 0, x if x >=0
    into
        (1 - slope) * Relu(x) + slope * x
    if
        1. PRelu's input[1] is not an initializer
    """

    pattern_to_match = ['PRelu']

    def pattern_matching(
        self, base_node: onnx.NodeProto  # pylint: disable=no-member
    ) -> Iterable[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            inits_to_add=self.make_new_init(matched_nodes),
            vis_to_add=self.make_new_vi(matched_nodes),
        )

        return base_node.input

    def pattern_condition_checker(
        self, nodes_to_check: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> bool:
        (prelu,) = nodes_to_check

        return prelu.input[1] not in self.initializer_map

    def make_new_node(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.NodeProto]:  # pylint: disable=no-member
        (prelu,) = matched_nodes

        relu = onnx.helper.make_node(
            'Relu',
            inputs=[prelu.input[0]],
            outputs=[prelu.output[0] + '_relu_out'],
            name=prelu.output[0] + '_0',
        )

        sub = onnx.helper.make_node(
            'Sub',
            inputs=[prelu.output[0] + '_ones_like_slope', prelu.input[1]],
            outputs=[prelu.output[0] + '_slope_fused'],
            name=prelu.output[0] + '_1',
        )

        mul_0 = onnx.helper.make_node(
            'Mul',
            inputs=[prelu.output[0] + '_relu_out', prelu.output[0] + '_slope_fused'],
            outputs=[prelu.output[0] + '_mul_0_out'],
            name=prelu.output[0] + '_2',
        )

        mul_1 = onnx.helper.make_node(
            'Mul',
            inputs=[prelu.input[0], prelu.input[1]],
            outputs=[prelu.output[0] + '_mul_1_out'],
            name=prelu.output[0] + '_3',
        )

        add = onnx.helper.make_node(
            'Add',
            inputs=[prelu.output[0] + '_mul_0_out', prelu.output[0] + '_mul_1_out'],
            outputs=prelu.output,
            name=prelu.output[0] + '_4',
        )

        return [relu, sub, mul_0, mul_1, add]

    def make_new_init(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.TensorProto]:  # pylint: disable=no-member
        (prelu,) = matched_nodes

        slope_shape = self.get_value_info_shape(prelu.input[1])
        slope_dtype = onnx.mapping.TENSOR_TYPE_TO_NP_TYPE[self.get_value_info_dtype(prelu.input[1])]

        ones_like_slope = np.ones(slope_shape).astype(slope_dtype)
        new_init = onnx.numpy_helper.from_array(
            ones_like_slope, prelu.output[0] + '_ones_like_slope'
        )

        return [new_init]

    def make_new_vi(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.ValueInfoProto]:  # pylint: disable=no-member
        (prelu,) = matched_nodes
        input_shape = self.get_value_info_shape(prelu.input[0])
        slope_shape = self.get_value_info_shape(prelu.input[1])
        input_dtype = self.get_value_info_dtype(prelu.input[0])

        return [
            onnx.helper.make_tensor_value_info(
                prelu.output[0] + '_relu_out',
                input_dtype,
                input_shape,
            ),
            onnx.helper.make_tensor_value_info(
                prelu.output[0] + '_slope_fused',
                input_dtype,
                slope_shape,
            ),
            onnx.helper.make_tensor_value_info(
                prelu.output[0] + '_mul_0_out',
                input_dtype,
                input_shape,
            ),
            onnx.helper.make_tensor_value_info(
                prelu.output[0] + '_mul_1_out',
                input_dtype,
                input_shape,
            ),
        ]



from typing import Dict, List, Optional, Sequence, Text, Tuple, Union

import numpy as np
import onnx
import onnxruntime as ort
import pytest

from furiosa.optimizer.frontend.onnx import __DOMAIN__, __OPSET_VERSION__


@pytest.fixture
def make_model():  # pylint: disable
    return _make_onnx_model_from_model_desc


@pytest.fixture
def make_test_model():
    def inner(
        opset_version: int,
        input_shape: Optional[Sequence[int]] = None,
        output_shape: Optional[Sequence[int]] = None,
    ) -> onnx.ModelProto:
        if input_shape is None:
            input_shape = [2, 8]
        if output_shape is None:
            output_shape = [2, 6]
        in_channel = input_shape[1]
        out_channel = output_shape[1]
        opsetid = ("", opset_version)
        return _make_onnx_model_from_model_desc(
            model_desc={
                "input": {"x": (np.float32, input_shape)},
                "output": {"y": (np.float32, output_shape)},
                "initializer": {
                    "w": (np.float32, [in_channel, out_channel]),
                    "a": (np.float32, [1, out_channel]),
                },
                "node": [
                    ("MatMul", ["x", "w"], ["0"]),
                    ("Add", ["0", "a"], ["y"]),
                ],
                "opsetid": [opsetid],
            }
        )

    return inner


def _make_onnx_model_from_model_desc(
    model_desc: Dict,
    doc_string: str = "",
    producer_name: str = "",
    check: bool = True,
    infer_shape: bool = True,
) -> onnx.ModelProto:  # pylint: disable=no-member
    """
    Type hints for model_desc

    model_desc = {
        "input": Dict[str, Tuple[np.dtype, List[int]]],
        "output": Dict[str, Tuple[np.dtype, List[int]]],
        "initializer": Dict[str, Union[Tuple[np.dtype, List[int]], np.ndarray]],
        "node": List[Tuple[str, List[str], List[str], Dict]],
        "opsetid": List[Tuple[str, str]],
    }
    """
    input_desc = model_desc["input"]
    output_desc = model_desc["output"]
    init_desc = model_desc.get("initializer", {})
    node_desc = model_desc.get("node", [])
    opsetid_desc = model_desc.get("opsetid", [(__DOMAIN__, __OPSET_VERSION__)])

    inputs = _make_value_info_list(input_desc)
    outputs = _make_value_info_list(output_desc)
    inits = _make_initializer_list(init_desc)
    nodes = _make_node_list(
        [desc if isinstance(desc[-1], dict) else (*desc, {}) for desc in node_desc]
    )
    opsetids = _make_opsetid_list(opsetid_desc)
    graph = _make_graph(nodes, inputs, outputs, inits, value_info=None, doc_string=doc_string)
    model = _make_model(graph, opsetids, check, producer_name)

    if infer_shape:
        model = onnx.shape_inference.infer_shapes(model)

    return model


def _make_value_info_list(
    vi_desc: Dict[str, Tuple[np.dtype, List[Union[Text, int]]]]
) -> List[onnx.ValueInfoProto]:  # pylint: disable=no-member
    return [
        onnx.helper.make_tensor_value_info(
            name, onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[np.dtype(np_dtype)], shape
        )
        for name, (np_dtype, shape) in vi_desc.items()
    ]


def _make_initializer_list(
    init_desc: Dict[str, Union[Tuple[np.dtype, List[Union[Text, int]]], np.ndarray]]
) -> List[onnx.TensorProto]:  # pylint: disable=no-member
    initailizers = []
    for name, arg in init_desc.items():
        if isinstance(arg, tuple):
            dtype, shape = arg
            initailizers.append(onnx.numpy_helper.from_array(_random_generator(dtype, shape), name))
        elif isinstance(arg, np.ndarray):
            np_arr = arg
            initailizers.append(onnx.numpy_helper.from_array(np_arr, name))
        else:
            raise TypeError(repr(type(arg)))

    return initailizers


def _make_node_list(
    node_desc: List[Tuple[str, List[str], List[str], Dict]]
) -> List[onnx.NodeProto]:  # pylint: disable=no-member
    return [
        onnx.helper.make_node(op_type, inputs, outputs, name=f"{op_type}_{idx}", **attrs)
        for idx, (op_type, inputs, outputs, attrs) in enumerate(node_desc)
    ]


def _make_opsetid_list(
    opsetids: List[Tuple[str, int]]
) -> List[onnx.OperatorSetIdProto]:  # pylint:disable=no-member
    return [onnx.helper.make_opsetid(domain, version) for (domain, version) in opsetids]


def _make_graph(
    nodes: List[onnx.NodeProto],  # pylint: disable=no-member
    inputs: List[onnx.ValueInfoProto],  # pylint: disable=no-member
    outputs: List[onnx.ValueInfoProto],  # pylint: disable=no-member
    initializer: Optional[List[onnx.TensorProto]] = None,  # pylint: disable=no-member
    value_info: Optional[List[onnx.ValueInfoProto]] = None,  # pylint: disable=no-member
    name: str = "graph",
    doc_string: str = "",
) -> onnx.GraphProto:  # pylint: disable=no-member
    graph_def = onnx.helper.make_graph(
        nodes, name, inputs, outputs, initializer, doc_string, value_info
    )
    return graph_def


def _make_model(
    graph: onnx.GraphProto,  # pylint: disable=no-member
    opset_imports: List[onnx.OperatorSetIdProto],  # pylint: disable=no-member
    check: bool = True,
    producer_name: str = "",
) -> onnx.ModelProto:  # pylint: disable=no-member
    model = onnx.helper.make_model(graph, opset_imports=opset_imports, producer_name=producer_name)
    if check:
        onnx.checker.check_model(model)
        ort.InferenceSession(model.SerializeToString())
    return model


def _random_generator(np_dtype, shape):
    rng = np.random.default_rng()
    if np.issubdtype(np_dtype, np.floating):
        return rng.standard_normal(shape, dtype=np.dtype(np_dtype))
    if np.issubdtype(np_dtype, np.integer):
        numeric_info = np.iinfo(np_dtype)
        return rng.integers(
            low=numeric_info.min,
            high=numeric_info.max,
            size=shape,
            dtype=np.dtype(np_dtype),
        )
    raise TypeError(repr(np_dtype))



import numpy as np

from furiosa.optimizer.frontend.onnx.transformer.utils import (
    eliminate_unused_input,
    eliminate_unused_output,
)


def test_eliminate_unused_input(make_model):
    input_shape = [4, 5]
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"x": (np.float32, input_shape)},
    }

    model = make_model(model_desc)
    model = eliminate_unused_input(model)

    # Graph input should not be removed if graph input = graph output
    assert model.graph.input


def test_eliminate_unused_output(make_model):
    input_shape = [1, 2, 3]
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"x": (np.float32, input_shape)},
    }

    model = make_model(model_desc)
    model = eliminate_unused_output(model)

    # Graph output should not be removed if graph input = graph output
    assert model.graph.output



from typing import Iterable, List, Optional

import numpy as np
import onnx
from onnx.helper import make_model
import onnxoptimizer

from furiosa.optimizer.frontend.onnx.transformer import ONNXTransformer, utils
from furiosa.optimizer.frontend.onnx.utils.inference_shape import InferenceShape
from furiosa.optimizer.frontend.onnx.utils.version_checker import CheckVersion
from furiosa.optimizer.interfaces.transformer import Transformer


class EmbeddingBagPorting(Transformer):
    # When https://github.com/pytorch/pytorch/issues/58567 issue gets resolved, this transformer
    # is no longer needed.
    def __init__(self, batch_size: Optional[int] = 128):
        self.batch_size = batch_size

    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:  # pylint: disable=no-member
        dynamic_axes = {"batch_size": self.batch_size}
        input_shapes = {
            input.name: [
                dynamic_axes.get(dim.dim_param, dim.dim_value)
                for dim in input.type.tensor_type.shape.dim
            ]
            for input in model.graph.input
        }
        model = CheckVersion().transform(model)
        model = utils.name_nodes(model)
        model = onnxoptimizer.optimize(model, passes=["extract_constant_to_initializer"])
        model = EmbeddingBagPattern(model).transform()
        model = InferenceShape(model).inference_shape(input_shapes)
        return model


class EmbeddingBagPattern(ONNXTransformer):
    """Apply to the DLRM model only.

    Transform onnx-exported EmbeddingBag graph to be equivalent to torch.EmbeddingBag
    If not applied on DLRM model, onnx simplifier doesn't work because transformed embeddingbag
    can't handle empty bag.

    https://github.com/pytorch/pytorch/blob/55d479aca5e959c5f2fe3089e162db710bb7632d/torch/onnx/symbolic_opset11.py#L836-L902
    """

    def pattern_matching(self, base_node):
        # pylint: disable=too-many-return-statements

        if base_node.op_type != 'Loop':
            return base_node.input
        subgraph = onnx.helper.get_attribute_value(base_node.attribute[0])
        subgraph_producer_map = {
            node_output: node for node in subgraph.node for node_output in node.output
        }
        subgraph_base_nodes = []
        for output in subgraph.output:
            output_producer = subgraph_producer_map[output.name]
            if output_producer.op_type == 'ReduceSum':
                subgraph_base_nodes.append(output_producer)
            elif output_producer.op_type in ['ReduceMean', 'ReduceMax']:
                # TODO case when EmbeddingBag mode is mean or max
                raise NotImplementedError(
                    "EmbeddingBagPorting of mean or max mode is not implemented"
                )

        if len(subgraph_base_nodes) != 1:
            return base_node.input

        subgraph_base_node = subgraph_base_nodes[0]
        try:
            gather_0 = subgraph_producer_map[subgraph_base_node.input[0]]
            if gather_0.op_type != "Gather":
                return base_node.input
            slice_0 = subgraph_producer_map[gather_0.input[1]]
            if slice_0.op_type != "Slice":
                return base_node.input
            unsqueeze_0 = subgraph_producer_map[slice_0.input[1]]
            if unsqueeze_0.op_type != "Unsqueeze":
                return base_node.input
            gather_1 = subgraph_producer_map[unsqueeze_0.input[0]]
            if gather_1.op_type != "Gather":
                return base_node.input
            unsqueeze_1 = subgraph_producer_map[slice_0.input[2]]
            if unsqueeze_1.op_type != "Unsqueeze":
                return base_node.input
            gather_2 = subgraph_producer_map[unsqueeze_1.input[0]]
            if gather_2.op_type != "Gather":
                return base_node.input
        except IndexError:
            return base_node.input

        matched_nodes = [
            subgraph_base_node,
            gather_0,
            slice_0,
            unsqueeze_0,
            gather_1,
            unsqueeze_1,
            gather_2,
        ]

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        subgraph_transformer = ONNXTransformer(make_model(subgraph), name_nodes=False)

        subgraph_transformer.transform_to_fuse(
            [subgraph_base_node],
            nodes_to_add=_make_subgraph_new_node(matched_nodes),
            inits_to_add=_make_subgraph_new_init(matched_nodes),
        )
        subgraph_transformer.build_optimized_model(subgraph_transformer.model, check=False)

        self.transform_to_fuse(
            [base_node],
            nodes_to_add=_make_new_node(base_node, subgraph_transformer.model.graph),
        )

        return base_node.input

    def pattern_condition_checker(
        self, nodes_to_check: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> bool:
        (
            subgraph_base_node,
            gather_0,
            slice_0,
            unsqueeze_0,
            gather_1,
            unsqueeze_1,
            gather_2,
        ) = nodes_to_check

        return (
            _check_condition_1(subgraph_base_node)
            and _check_condition_2(gather_0)
            and self.check_condition_3(slice_0)
            and _check_condition_4(unsqueeze_0)
            and _check_condition_4(unsqueeze_1)
            and _check_condition_5(gather_1, gather_2)
        )

    def check_condition_3(self, node: onnx.NodeProto) -> bool:  # pylint: disable=no-member
        axes = self.get_initializer_array(node.input[3])

        return np.array_equal(axes, [0])


def _check_condition_1(node: onnx.NodeProto) -> bool:  # pylint: disable=no-member
    attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}
    axes = attrs.get("axes")
    keepdims = attrs.get("keepdims", 1)

    return axes is not None and axes == [0] and keepdims == 0


def _check_condition_2(node: onnx.NodeProto) -> bool:  # pylint: disable=no-member
    attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}
    axis = attrs.get("axis", 0)

    return axis == 0


def _check_condition_4(node: onnx.NodeProto) -> bool:  # pylint: disable=no-member
    attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}
    axes = attrs["axes"]

    return axes == [0]


def _check_condition_5(
    node_1: onnx.NodeProto, node_2: onnx.NodeProto  # pylint: disable=no-member
) -> bool:
    for node in [node_1, node_2]:
        attrs = {attr.name: onnx.helper.get_attribute_value(attr) for attr in node.attribute}
        axis = attrs.get("axis", 0)
        if axis != 0:
            return False
    return node_1.input[1] == node_2.input[1]


def _make_subgraph_new_init(
    matched_nodes: Iterable[onnx.NodeProto],  # pylint: disable=no-member
) -> List[onnx.TensorProto]:  # pylint: disable=no-member
    _, gather, *_ = matched_nodes
    return [
        onnx.numpy_helper.from_array(np.array([1]), gather.output[0] + '_indice_1'),
    ]


def _make_subgraph_new_node(
    matched_nodes: Iterable[onnx.NodeProto],  # pylint: disable=no-member
) -> List[onnx.NodeProto]:  # pylint: disable=no-member
    subgraph_base_node, gather, *_ = matched_nodes
    return [
        onnx.helper.make_node(
            'Shape',
            inputs=[gather.output[0]],
            outputs=[gather.output[0] + '_shape'],
            name=gather.name + '_shape',
        ),
        onnx.helper.make_node(
            'Gather',
            inputs=[gather.output[0] + '_shape', gather.output[0] + '_indice_1'],
            outputs=[gather.output[0] + '_shape_1'],
            name=gather.name + '_get_shape_1',
        ),
        onnx.helper.make_node(
            'ConstantOfShape',
            inputs=[gather.output[0] + '_shape_1'],
            outputs=[gather.output[0] + '_zero_vec'],
            name=gather.name + '_zero_vec',
        ),
        onnx.helper.make_node(
            'Unsqueeze',
            inputs=[gather.output[0] + '_zero_vec'],
            outputs=[gather.output[0] + '_zero_vec_unsqueezed'],
            name=gather.name + '_zero_vec_unsqueeze',
            axes=[0],
        ),
        onnx.helper.make_node(
            'Concat',
            inputs=[gather.output[0], gather.output[0] + '_zero_vec_unsqueezed'],
            outputs=[gather.output[0] + '_nonzero'],
            name=gather.name + '_convert',
            axis=0,
        ),
        onnx.helper.make_node(
            subgraph_base_node.op_type,
            inputs=[gather.output[0] + '_nonzero'],
            outputs=[subgraph_base_node.output[0]],
            name=subgraph_base_node.name,
            **{
                attr.name: onnx.helper.get_attribute_value(attr)
                for attr in subgraph_base_node.attribute
            },
        ),
    ]


def _make_new_node(
    base_node: onnx.NodeProto, subgraph: onnx.GraphProto  # pylint: disable=no-member
) -> List[onnx.NodeProto]:  # pylint: disable=no-member
    return [
        onnx.helper.make_node(
            base_node.op_type,
            inputs=base_node.input,
            outputs=base_node.output,
            name=base_node.name,
            body=subgraph,
        )
    ]



import numpy as np

from furiosa.optimizer.frontend.onnx.transformer.convert_negative_pads_to_slice import (
    ConvertNegativePadsToSlice,
)
from furiosa.optimizer.frontend.onnx.transformer.utils import check_value_info


def test_case1(make_transformer_model, check_output_value, check_graph_node):
    input_shape = [1, 2, 4, 4]
    output_shape = [1, 2, 4, 4]
    opsetid = ("", 13)
    model_desc = {
        "input": {"data": (np.float32, input_shape)},
        "output": {"output": (np.float32, output_shape)},
        "initializer": {
            "pads": np.array([0, 0, -1, -1, 0, 0, 1, 1], dtype=np.int64),
            "constant_value": np.array([0], dtype=np.float32),
        },
        "node": [
            ("Pad", ["data", "pads", "constant_value"], ["output"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = make_transformer_model(model_desc, ConvertNegativePadsToSlice())

    check_output_value(orig_model, trans_model, [input_shape])
    check_graph_node(trans_model, op_types=["Slice", "Pad"])
    check_value_info(trans_model)


def test_case2(make_transformer_model, check_output_value, check_graph_node):
    input_shape = [1, 2, 4, 4]
    output_shape = [1, 2, 2, 1]
    opsetid = ("", 13)
    model_desc = {
        "input": {"data": (np.float32, input_shape)},
        "output": {"output": (np.float32, output_shape)},
        "initializer": {
            "pads": np.array([0, 0, -1, -1, 0, 0, -1, -2], dtype=np.int64),
            "constant_value": np.array([0], dtype=np.float32),
        },
        "node": [
            ("Pad", ["data", "pads", "constant_value"], ["output"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = make_transformer_model(model_desc, ConvertNegativePadsToSlice())

    check_output_value(orig_model, trans_model, [input_shape])
    check_graph_node(trans_model, op_types=["Slice"])
    check_value_info(trans_model)


def test_case3(make_transformer_model, check_output_value, check_graph_node):
    input_shape = [1, 2, 4, 4]
    output_shape = [1, 2, 0, 4]
    opsetid = ("", 13)
    model_desc = {
        "input": {"data": (np.float32, input_shape)},
        "output": {"output": (np.float32, output_shape)},
        "initializer": {
            "pads": np.array([0, 0, -2, -1, 0, 0, -2, 1], dtype=np.int64),
            "constant_value": np.array([0], dtype=np.float32),
        },
        "node": [
            ("Pad", ["data", "pads", "constant_value"], ["output"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = make_transformer_model(model_desc, ConvertNegativePadsToSlice())

    check_output_value(orig_model, trans_model, [input_shape])
    check_graph_node(trans_model, op_types=["Slice", "Pad"])
    check_value_info(trans_model)


def test_case4(make_transformer_model, check_output_value, check_graph_node):
    input_shape = [1, 2, 4, 4]
    output_shape = [1, 2, 1, 5]
    opsetid = ("", 13)
    model_desc = {
        "input": {"data": (np.float32, input_shape)},
        "output": {"output": (np.float32, output_shape)},
        "initializer": {
            "pads": np.array([0, 0, -1, -2, 0, 0, -2, 3], dtype=np.int64),
            "constant_value": np.array([0], dtype=np.float32),
        },
        "node": [
            ("Pad", ["data", "pads", "constant_value"], ["output"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = make_transformer_model(model_desc, ConvertNegativePadsToSlice())

    check_output_value(orig_model, trans_model, [input_shape])
    check_graph_node(trans_model, op_types=["Slice", "Pad"])
    check_value_info(trans_model)


def test_case5(make_transformer_model, check_graph_node):
    input_shape = [1, 2, 4, 4]
    output_shape = [1, 2, 4, 4]
    opsetid = ("", 13)
    model_desc = {
        "input": {"data": (np.float32, input_shape)},
        "output": {"output": (np.float32, output_shape)},
        "initializer": {
            "pads": np.array([0, 0, -4, -1, 0, 0, 4, 1], dtype=np.int64),
            "constant_value": np.array([0], dtype=np.float32),
        },
        "node": [
            ("Pad", ["data", "pads", "constant_value"], ["output"]),
        ],
        "opsetid": [opsetid],
    }

    _orig_model, trans_model = make_transformer_model(model_desc, ConvertNegativePadsToSlice())
    # Don't check output value, but check that transformation is not happened.
    # Though orig_model and trans_model are same, they output 'nan' intermittently,
    # which leads to failure of check_output_value.
    check_graph_node(trans_model, op_types=["Pad"])
    check_value_info(trans_model)



import numpy as np

from furiosa.optimizer.frontend.onnx.transformer.fuse_gather_matmul import Pattern_1
from furiosa.optimizer.frontend.onnx.transformer.utils import check_value_info


def test_case1(make_transformer_model, check_output_value, check_graph_node):
    in_dims = 3
    num_embs = 10
    emb_dims = 128
    out_dims = 8

    model_desc = {
        "input": {"x": (np.int64, [in_dims])},
        "output": {"y": (np.float32, [in_dims, out_dims])},
        "initializer": {
            "table": (np.float32, [num_embs, emb_dims]),
            "w": (np.float32, [emb_dims, out_dims]),
        },
        "node": [
            ("Gather", ["table", "x"], ["0"]),
            ("MatMul", ["0", "w"], ["y"]),
        ],
    }
    orig_model, trans_model = make_transformer_model(model_desc, Pattern_1)

    check_graph_node(trans_model, op_types=["Gather"])
    check_output_value(
        orig_model,
        trans_model,
        [(in_dims)],
        data=[np.random.default_rng().integers(0, num_embs, size=in_dims)],
    )
    check_value_info(trans_model)


def test_case2(make_transformer_model, check_output_value, check_graph_node):
    seq_len = 2
    in_dims = 3
    num_embs = 10
    emb_dims = 128
    out_dims = 8

    model_desc = {
        "input": {
            "x": (np.int64, [seq_len, in_dims]),
        },
        "output": {"y": (np.float32, [seq_len, in_dims, out_dims])},
        "initializer": {
            "table": (np.float32, [num_embs, emb_dims]),
            "w": (np.float32, [emb_dims, out_dims]),
        },
        "node": [
            ("Gather", ["table", "x"], ["0"]),
            ("MatMul", ["0", "w"], ["y"]),
        ],
    }
    orig_model, trans_model = make_transformer_model(model_desc, Pattern_1)

    check_graph_node(trans_model, op_types=["Gather"])
    check_output_value(
        orig_model,
        trans_model,
        [(seq_len, in_dims)],
        data=[np.random.default_rng().integers(0, num_embs, size=(seq_len, in_dims))],
    )
    check_value_info(trans_model)


def test_case3(make_transformer_model, check_output_value, check_graph_node):
    """
    Test RHS of Pattern_1 condition 7
    """
    in_dims = 3
    num_embs = 10
    emb_dims = 128
    out_dims = 8

    model_desc = {
        "input": {"x": (np.int64, [in_dims])},
        "output": {"y": (np.float32, [out_dims, in_dims])},
        "initializer": {
            "table": (np.float32, [emb_dims, num_embs]),
            "w": (np.float32, [out_dims, emb_dims]),
        },
        "node": [
            ("Gather", ["table", "x"], ["0"], {"axis": 1}),
            ("MatMul", ["w", "0"], ["y"]),
        ],
    }
    orig_model, trans_model = make_transformer_model(model_desc, Pattern_1)

    check_graph_node(trans_model, op_types=["Gather"])
    check_output_value(
        orig_model,
        trans_model,
        [(in_dims)],
        data=[np.random.default_rng().integers(0, num_embs, size=in_dims)],
    )
    check_value_info(trans_model)


def test_case4(make_transformer_model, check_graph_node):
    """
    Test case which does not meet condition 1
    """
    in_dims = 3
    num_embs = 10
    emb_dims = 128
    out_dims = 8

    model_desc = {
        "input": {
            "x": (np.int64, [in_dims]),
            "w": (np.float32, [emb_dims, out_dims]),
        },
        "output": {"y": (np.float32, [in_dims, out_dims])},
        "initializer": {
            "table": (np.float32, [num_embs, emb_dims]),
        },
        "node": [
            ("Gather", ["table", "x"], ["0"]),
            ("MatMul", ["0", "w"], ["y"]),
        ],
    }
    _, trans_model = make_transformer_model(model_desc, Pattern_1)

    # fusion does not apply
    check_graph_node(trans_model, op_types=["Gather", "MatMul"])
    check_value_info(trans_model)


def test_case5(make_transformer_model, check_graph_node):
    """
    Test case which does not meet condition 2
    """
    in_dims = 16
    num_embs = 8
    emb_dims = 256
    out_dims = 4

    model_desc = {
        "input": {
            "x": (np.int64, [in_dims]),
            "table": (np.float32, [num_embs, emb_dims]),
        },
        "output": {"y": (np.float32, [in_dims, out_dims])},
        "initializer": {
            "w": (np.float32, [emb_dims, out_dims]),
        },
        "node": [
            ("Gather", ["table", "x"], ["0"]),
            ("MatMul", ["0", "w"], ["y"]),
        ],
    }
    _, trans_model = make_transformer_model(model_desc, Pattern_1)

    # fusion does not apply
    check_graph_node(trans_model, op_types=["Gather", "MatMul"])
    check_value_info(trans_model)


def test_case6(make_transformer_model, check_graph_node):
    """
    Test case which does not meet condition 3 & 5
    """
    in_dims = 16
    num_embs = 8
    emb_dims = 256
    out_dims = 4

    model_desc = {
        "input": {"x": (np.int64, [in_dims])},
        "output": {"y": (np.int32, [in_dims, out_dims])},
        "initializer": {
            "table": (np.int32, [num_embs, emb_dims]),
            "w": (np.int32, [emb_dims, out_dims]),
        },
        "node": [
            ("Gather", ["table", "x"], ["0"]),
            ("MatMul", ["0", "w"], ["y"]),
        ],
    }
    _, trans_model = make_transformer_model(model_desc, Pattern_1)

    # fusion does not apply
    check_graph_node(trans_model, op_types=["Gather", "MatMul"])
    check_value_info(trans_model)


def test_case7(make_transformer_model, check_graph_node):
    """
    Test case which does not meet condition 4
    """
    seq_len = 3
    in_dims = 32
    out_dims = 4

    model_desc = {
        "input": {"x": (np.int64, [seq_len, in_dims])},
        "output": {"y": (np.int32, [seq_len, out_dims])},
        "initializer": {
            "table": (np.int32, [in_dims]),
            "w": (np.int32, [in_dims, out_dims]),
        },
        "node": [
            ("Gather", ["table", "x"], ["0"]),
            ("MatMul", ["0", "w"], ["y"]),
        ],
    }
    _, trans_model = make_transformer_model(model_desc, Pattern_1)

    # fusion does not apply
    check_graph_node(trans_model, op_types=["Gather", "MatMul"])
    check_value_info(trans_model)



from typing import Dict, List, Mapping, Optional, Tuple

import numpy as np
import onnx

from furiosa.optimizer.frontend.onnx.transformer.infer_squeeze_axes import InferSqueezeAxes
from furiosa.optimizer.frontend.onnx.transformer.utils import check_value_info, fixed_point
from furiosa.optimizer.frontend.onnx.utils.inference_shape import InferenceShape


def test_case1(make_model, check_output_value):
    input_shape = [1, 4, 1, 1]
    output_shape = [8]
    opsetid = ("", 12)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [4, 8]),
            "b": (np.float32, [8]),
        },
        "node": [
            ("Squeeze", ["x"], ["x_1"]),
            ("MatMul", ["x_1", "w"], ["z"]),
            ("Add", ["z", "b"], ["y"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = _make_test_model(make_model, model_desc, {"x": input_shape})

    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)
    assert [0, 2, 3] == trans_model.graph.node[0].attribute[0].ints


def test_case2(make_model, check_output_value):
    input_shape = [1, 4, 1, 1]
    output_shape = [4]
    opsetid = ("", 12)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {},
        "node": [
            ("Squeeze", ["x"], ["x_1"]),
            ("Unsqueeze", ["x_1"], ["x_2"], {"axes": [0, 2, 3]}),
            ("Squeeze", ["x_2"], ["x_3"]),
            ("Unsqueeze", ["x_3"], ["x_4"], {"axes": [0, 2, 3]}),
            ("Squeeze", ["x_4"], ["y"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = _make_test_model(make_model, model_desc, {"x": input_shape})

    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)
    assert [0, 2, 3] == trans_model.graph.node[0].attribute[0].ints
    assert [0, 2, 3] == trans_model.graph.node[2].attribute[0].ints
    assert [0, 2, 3] == trans_model.graph.node[4].attribute[0].ints


def _make_test_model(
    make_model, model_desc: Dict, input_shapes: Mapping[str, List[Optional[int]]]
) -> Tuple[onnx.ModelProto, onnx.ModelProto]:  # pylint: disable=no-member
    orig_model = make_model(model_desc)
    trans_model = fixed_point(
        orig_model,
        [
            lambda model: InferenceShape(model).inference_shape(input_shapes),
            InferSqueezeAxes().transform,
        ],
    )

    return orig_model, trans_model



from typing import Iterable, List, Mapping, Union

import numpy as np
import onnx

from furiosa.optimizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.optimizer.interfaces.transformer import Transformer


class ConvertNegativePadsToSlice(Transformer):
    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:  # pylint: disable=no-member
        # Pattern_1 - Pattern_2 order should be preserved
        # in order to remove noop-like 'Pad' operator generated by Pattern_1.
        for transformer in [
            Pattern_1,
            Pattern_2,
        ]:
            model = transformer(model).transform()

        return model


class Pattern_1(ONNXTransformer):
    """
    transform
        prev --> Pad --> next
    to
        prev --> Slice --> Pad --> next
    if
        1. Pad's pads input (pad.input[1]) is an initializer
        2. at least one of pads value is negative
        3. Sum of negative pads value in axis i does not exceed corresponding input_shape[i]
          (if not, model is invalid)
        4. absolute value of each negative pads in axis i is less than corresponding input_shape[i]
          (if not, leads to invalid model/'nan' output)
    """

    pattern_to_match = ['Pad']

    def pattern_matching(
        self, base_node: onnx.NodeProto  # pylint: disable=no-member
    ) -> Iterable[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        self.transform_to_fuse(
            matched_nodes,
            nodes_to_add=self.make_new_node(matched_nodes),
            **self.make_new_init_and_vi(matched_nodes),  # type: ignore
        )

        return base_node.input

    def pattern_condition_checker(
        self, nodes_to_check: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> bool:
        (pad,) = nodes_to_check
        if pad.input[1] not in self.initializer_map:
            return False

        pads = self.get_initializer_array(pad.input[1])
        if not any(pad < 0 for pad in pads):
            return False

        input_shape = self.get_value_info_shape(pad.input[0])
        pad_per_axis = zip(pads[: len(pads) // 2], pads[len(pads) // 2 :])

        # If sum of negative pad value exceeds input size, output size is negative, which leads to
        # invaiid Pad operator>
        # If one of negative pad value equals input size, original onnx Pad outputs 'nan'
        # intermittently.
        # If one of negative pad value exceeds input size, Pad is invalid
        return all(
            input_size + pad_begin + pad_end >= 0
            and input_size + pad_begin > 0
            and input_size + pad_end > 0
            for (input_size, (pad_begin, pad_end)) in zip(input_shape, pad_per_axis)
        )

    def make_new_node(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> List[onnx.NodeProto]:  # pylint: disable=no-member
        (pad,) = matched_nodes

        slice_ = onnx.helper.make_node(
            'Slice',
            inputs=[
                pad.input[0],
                pad.output[0] + '_starts',
                pad.output[0] + '_ends',
                pad.output[0] + '_axes',
            ],
            outputs=[pad.output[0] + '_slice_out'],
            name=pad.output[0] + '_0',
        )

        new_pad = onnx.helper.make_node(
            'Pad',
            inputs=[pad.output[0] + '_slice_out', pad.output[0] + '_pads', *pad.input[2:]],
            outputs=pad.output,
            name=pad.output[0] + '_1',
        )

        return [slice_, new_pad]

    def make_new_init_and_vi(
        self, matched_nodes: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> Mapping[
        str, Union[List[onnx.NodeProto], List[onnx.ValueInfoProto], List[onnx.TensorProto]]
    ]:  # pylint: disable=no-member
        (pad,) = matched_nodes
        input_shape = self.get_value_info_shape(pad.input[0])
        new_inits = []

        pads = self.get_initializer_array(pad.input[1])
        pad_per_axis = list(zip(pads[: len(pads) // 2], pads[len(pads) // 2 :]))
        starts, ends, axes = zip(
            *[
                (max(-pad_begin, 0), input_shape[i] + min(pad_end, 0), i)
                for (i, (pad_begin, pad_end)) in enumerate(pad_per_axis)
                if pad_begin < 0 or pad_end < 0
            ]
        )
        pads = [max(pad, 0) for pad in pads]
        for init_list, suffix in (
            (starts, '_starts'),
            (ends, '_ends'),
            (axes, '_axes'),
            (pads, '_pads'),
        ):
            new_init = onnx.numpy_helper.from_array(
                np.array(init_list, dtype=np.int64), pad.output[0] + suffix
            )
            new_inits.append(new_init)

        new_vis = []
        slice_output_shape = [
            int(shape + min(pad_begin, 0) + min(pad_end, 0))
            for (shape, (pad_begin, pad_end)) in zip(input_shape, pad_per_axis)
        ]
        slice_output_dtype = self.get_value_info_dtype(pad.output[0])
        slice_output_vi = onnx.helper.make_tensor_value_info(
            pad.output[0] + '_slice_out',
            slice_output_dtype,
            slice_output_shape,
        )
        new_vis.append(slice_output_vi)

        return {'inits_to_add': new_inits, 'vis_to_add': new_vis}


class Pattern_2(ONNXTransformer):
    """
    transform
        prev --> Pad --> next
    to
        prev --> next
    if
        1. Pad's pads value are all zero
    """

    pattern_to_match = ['Pad']

    def pattern_matching(
        self, base_node: onnx.NodeProto  # pylint: disable=no-member
    ) -> Iterable[str]:
        matched_nodes = self.pattern_matcher(base_node, self.pattern_to_match)
        if not matched_nodes:
            return base_node.input

        if not self.pattern_condition_checker(matched_nodes):
            return base_node.input

        (pad,) = matched_nodes

        self.transform_to_eliminate(matched_nodes, pad.input[0])

        return base_node.input

    def pattern_condition_checker(
        self, nodes_to_check: Iterable[onnx.NodeProto]  # pylint: disable=no-member
    ) -> bool:
        (pad,) = nodes_to_check
        if pad.input[1] not in self.initializer_map:
            return False

        pads = self.get_initializer_array(pad.input[1])
        return all(pad == 0 for pad in pads)



import numpy as np

from furiosa.optimizer.frontend.onnx.transformer.fuse_batchnorm import (
    Pattern_1,
    Pattern_2,
    Pattern_3,
    Pattern_4,
)
from furiosa.optimizer.frontend.onnx.transformer.utils import check_value_info


def test_case1(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 3
    input_shape = [1, in_channel, 4, 4]
    out_channel = 4
    output_shape = [1, out_channel, 3, 3]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)

    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [out_channel, in_channel, 2, 2]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
        },
        "node": [
            ("Conv", ["x", "w", "b"], ["0"]),
            ("BatchNormalization", ["0", "beta", "gamma", "mu", "var"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_1)

    check_graph_node(trans_model, op_types=["Conv"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case2(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 3
    input_shape = [1, in_channel, 4, 4]
    out_channel = 4
    output_shape = [1, out_channel, 3, 3]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)

    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [out_channel, in_channel, 2, 2]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
        },
        "node": [
            ("Conv", ["x", "w", "b"], ["0"]),
            ("Relu", ["0"], ["1"]),
            ("BatchNormalization", ["1", "beta", "gamma", "mu", "var"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_4)

    check_graph_node(trans_model, op_types=["Conv", "Relu", "Mul", "Add"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case3(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 4
    input_shape = [1, in_channel, 4, 4]
    out_channel = 8
    output_shape = [1, out_channel, 2, 2]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w1": (np.float32, [1, in_channel]),
            "w": (np.float32, [out_channel, in_channel, 3, 3]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
            "a": (np.float32, output_shape),
        },
        "node": [
            ("Mul", ["x", "w1"], ["0"]),
            ("Conv", ["0", "w", "b"], ["1"]),
            ("BatchNormalization", ["1", "beta", "gamma", "mu", "var"], ["2"]),
            ("Add", ["2", "a"], ["y"]),
        ],
    }
    orig_model, trans_model = make_transformer_model(model_desc, Pattern_1)

    check_graph_node(trans_model, op_types=["Mul", "Conv", "Add"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case4(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 5
    input_shape = [1, in_channel, 4, 4]
    out_channel = 2
    output_shape = [1, out_channel, 2, 2]

    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)

    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w1": (np.float32, input_shape),
            "w": (np.float32, [out_channel, in_channel, 3, 3]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
            "a": (np.float32, output_shape),
        },
        "node": [
            ("Mul", ["x", "w1"], ["0"]),
            ("Conv", ["0", "w", "b"], ["1"]),
            ("Relu", ["1"], ["2"]),
            ("BatchNormalization", ["2", "beta", "gamma", "mu", "var"], ["3"]),
            ("Add", ["a", "3"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_4)

    check_graph_node(trans_model, op_types=["Mul", "Conv", "Relu", "Mul", "Add", "Add"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case5(make_transformer_model, check_output_value, check_graph_node):
    """
    This tests Pattern_3
    """
    in_channel = 3
    input_shape = [1, in_channel, 8, 8]
    out_channel = 4
    output_shape = [1, out_channel, 6, 6]

    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [out_channel, in_channel, 3, 3]),
            "b": (np.float32, [out_channel]),
            "w1": (np.float32, [1, out_channel, 1, 1]),
            "a": (np.float32, [1, out_channel, 1, 1]),
        },
        "node": [
            ("Conv", ["x", "w", "b"], ["0"]),
            ("Mul", ["0", "w1"], ["1"]),
            ("Add", ["1", "a"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_3)

    check_graph_node(trans_model, op_types=["Conv"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case5_1(make_transformer_model, check_output_value, check_graph_node):
    """
    This tests Pattern_3
    """
    in_channel = 3
    input_shape = [1, in_channel, 8, 8]
    out_channel = 4
    output_shape = [1, out_channel, 6, 6]

    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [out_channel, in_channel, 3, 3]),
            "b": (np.float32, [out_channel]),
        },
        "node": [
            ("Conv", ["x", "w", "b"], ["0"]),
            ("Mul", ["0", "0"], ["1"]),
            ("Add", ["1", "1"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_3)

    check_graph_node(trans_model, op_types=["Conv", "Mul", "Add"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case5_2(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 4
    input_shape = [1, in_channel, 4, 4]
    out_channel = 4
    output_shape = [1, out_channel, 2, 2]

    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [out_channel, in_channel, 3, 3]),
            "w1": (np.float32, [1, out_channel, 2, 2]),
            "b": (np.float32, [out_channel]),
        },
        "node": [
            ("Conv", ["x", "w", "b"], ["0"]),
            ("Mul", ["0", "w1"], ["1"]),
            ("Add", ["1", "1"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_3)

    check_graph_node(trans_model, op_types=["Conv", "Mul", "Add"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case6(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 3
    input_shape = [1, in_channel, 4, 4]
    out_channel = 4
    output_shape = [1, out_channel, 5, 5]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [in_channel, out_channel, 2, 2]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
        },
        "node": [
            ("ConvTranspose", ["x", "w", "b"], ["0"]),
            ("BatchNormalization", ["0", "beta", "gamma", "mu", "var"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_2)

    check_graph_node(trans_model, op_types=["ConvTranspose"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case7(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 3
    input_shape = [1, in_channel, 4, 4]
    out_channel = 4
    output_shape = [1, out_channel, 5, 5]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [in_channel, out_channel, 2, 2]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
        },
        "node": [
            ("ConvTranspose", ["x", "w", "b"], ["0"]),
            ("Relu", ["0"], ["1"]),
            ("BatchNormalization", ["1", "beta", "gamma", "mu", "var"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_2)

    check_graph_node(trans_model, op_types=["ConvTranspose", "Relu", "BatchNormalization"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case8(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 4
    input_shape = [1, in_channel, 4, 4]
    out_channel = 8
    output_shape = [1, out_channel, 5, 5]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w1": (np.float32, input_shape),
            "w": (np.float32, [in_channel, out_channel, 2, 2]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
            "a": (np.float32, output_shape),
        },
        "node": [
            ("Mul", ["x", "w1"], ["0"]),
            ("ConvTranspose", ["0", "w", "b"], ["1"]),
            ("BatchNormalization", ["1", "beta", "gamma", "mu", "var"], ["2"]),
            ("Add", ["2", "a"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_2)

    check_graph_node(trans_model, op_types=["Mul", "ConvTranspose", "Add"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case9(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 5
    input_shape = [1, in_channel, 4, 4]
    out_channel = 2
    output_shape = [1, out_channel, 5, 5]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w1": (np.float32, input_shape),
            "w": (np.float32, [in_channel, out_channel, 2, 2]),
            "b": (np.float32, [out_channel]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
            "a": (np.float32, output_shape),
        },
        "node": [
            ("Mul", ["x", "w1"], ["1"]),
            ("ConvTranspose", ["1", "w", "b"], ["2"]),
            ("Relu", ["2"], ["3"]),
            ("BatchNormalization", ["3", "beta", "gamma", "mu", "var"], ["4"]),
            ("Add", ["4", "a"], ["y"]),
        ],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_2)

    check_graph_node(
        trans_model,
        op_types=["Mul", "ConvTranspose", "Relu", "BatchNormalization", "Add"],
    )
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case10(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 16
    input_shape = [1, in_channel, 6, 6]
    out_channel = 2
    output_shape = [1, out_channel, 3, 3]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [out_channel, in_channel, 4, 4]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
        },
        "node": [
            (
                "Conv",
                ["x", "w"],
                ["1"],
                {
                    "dilations": [1, 1],
                    "group": 1,
                    "kernel_shape": [4, 4],
                    "pads": [1, 1, 1, 1],
                    "strides": [2, 2],
                },
            ),
            ("BatchNormalization", ["1", "beta", "gamma", "mu", "var"], ["y"]),
        ],
    }
    orig_model, trans_model = make_transformer_model(model_desc, Pattern_1)

    check_graph_node(trans_model, op_types=["Conv"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def test_case11(make_transformer_model, check_output_value, check_graph_node):
    in_channel = 16
    input_shape = [1, in_channel, 4, 4]
    out_channel = 2
    output_shape = [1, out_channel, 8, 8]
    scale, B, input_mean, input_var = _bn_param_generator(num_features=out_channel)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {
            "w": (np.float32, [in_channel, out_channel, 4, 4]),
            "beta": scale,
            "gamma": B,
            "mu": input_mean,
            "var": input_var,
        },
        "node": [
            (
                "ConvTranspose",
                ["x", "w"],
                ["1"],
                {
                    "dilations": [1, 1],
                    "group": 1,
                    "kernel_shape": [4, 4],
                    "pads": [1, 1, 1, 1],
                    "strides": [2, 2],
                },
            ),
            ("BatchNormalization", ["1", "beta", "gamma", "mu", "var"], ["y"]),
        ],
    }
    orig_model, trans_model = make_transformer_model(model_desc, Pattern_2)

    check_graph_node(trans_model, op_types=["ConvTranspose"])
    check_output_value(orig_model, trans_model, [input_shape])
    check_value_info(trans_model)


def _bn_param_generator(num_features):
    """
    returns scale, B, input_mean and input_var
    as defined in https://github.com/onnx/onnx/blob/master/docs/Operators.md#inputs-12
    """
    rng = np.random.default_rng()
    return (
        rng.standard_normal(num_features, dtype=np.float32),
        rng.standard_normal(num_features, dtype=np.float32),
        rng.standard_normal(num_features, dtype=np.float32),
        rng.standard_normal(num_features, dtype=np.float32) ** 2,
    )



import onnxoptimizer

from furiosa.optimizer.frontend.onnx.utils.inference_shape import __SKIPPED_PASSES__

# Allowed passes with following package versions
# onnx-simplifier   0.4.33
# onnxoptimizer     0.3.13
__ALLOWED_PASSES__ = set(
    [
        "eliminate_common_subexpression",
        "eliminate_deadend",
        "eliminate_identity",
        "eliminate_if_with_const_cond",
        "eliminate_nop_cast",
        "eliminate_nop_concat",
        "eliminate_nop_dropout",
        "eliminate_nop_expand",
        "eliminate_nop_flatten",
        "eliminate_nop_monotone_argmax",
        "eliminate_nop_pad",
        "eliminate_nop_reshape",
        "eliminate_nop_split",
        "eliminate_nop_transpose",
        "eliminate_nop_with_unit",
        "eliminate_shape_gather",
        "eliminate_shape_op",
        "eliminate_slice_after_shape",
        "eliminate_unused_initializer",
        "extract_constant_to_initializer",
        "fuse_concat_into_reshape",
        "fuse_consecutive_concats",
        "fuse_consecutive_log_softmax",
        "fuse_consecutive_reduce_unsqueeze",
        "fuse_consecutive_slices",
        "fuse_consecutive_squeezes",
        "fuse_consecutive_transposes",
        "fuse_consecutive_unsqueezes",
        "fuse_matmul_add_bias_into_gemm",
        "fuse_pad_into_conv",
        "fuse_pad_into_pool",
        "fuse_transpose_into_gemm",
        "nop",
    ]
)


def test_onnx_optimizer_pass():
    added_passes = (
        set(onnxoptimizer.get_fuse_and_elimination_passes())
        .difference(__SKIPPED_PASSES__)
        .difference(__ALLOWED_PASSES__)
    )

    assert not added_passes, f"New onnx optimizer passes are found {added_passes}"



from typing import Callable, List, Mapping, Optional, Sequence

import onnx

__DOMAIN__ = ''
__OPSET_VERSION__ = 13

from furiosa.optimizer.frontend.onnx.transformer.convert_negative_pads_to_slice import (
    ConvertNegativePadsToSlice,
)
from furiosa.optimizer.frontend.onnx.transformer.convert_prelu_to_relu import ConvertPReluToRelu
from furiosa.optimizer.frontend.onnx.transformer.fuse_batchnorm import FuseBatchNorm
from furiosa.optimizer.frontend.onnx.transformer.fuse_gather_matmul import FuseGatherMatMul
from furiosa.optimizer.frontend.onnx.transformer.polish_model import PolishModel
from furiosa.optimizer.frontend.onnx.utils.version_checker import CheckVersion


def optimize_model(
    model: onnx.ModelProto,  # pylint: disable=no-member
    input_shapes: Optional[Mapping[str, List[Optional[int]]]] = None,
    opset_version: int = __OPSET_VERSION__,
) -> onnx.ModelProto:  # pylint: disable=no-member
    model = _transform([CheckVersion(opset_version).transform], model)
    model = _transform([PolishModel(input_shapes).transform], model)

    # TODO check if graph_transform should apply.
    model = _transform([_reify], model)
    return model


def _transform(
    transformers: Sequence[
        Callable[[onnx.ModelProto], onnx.ModelProto]
    ],  # pylint: disable=no-member
    model: onnx.ModelProto,  # pylint: disable=no-member
) -> onnx.ModelProto:  # pylint: disable=no-member
    for transform in transformers:
        model = transform(model)
    return model


def _reify(model: onnx.ModelProto) -> onnx.ModelProto:  # pylint: disable=no-member
    transformers = [
        ConvertNegativePadsToSlice().transform,
        FuseBatchNorm().transform,
        FuseGatherMatMul().transform,
        ConvertPReluToRelu().transform,
    ]
    return _transform(transformers, model)



from typing import Dict, Iterable, List, Optional, Sequence, Type, Union

import numpy as np
import onnx
import onnxruntime as ort
import pytest

from furiosa.optimizer.frontend.onnx.transformer import ONNXTransformer
from furiosa.optimizer.interfaces.transformer import Transformer


@pytest.fixture
def make_transformer_model(make_model):
    def inner(
        model_desc: Dict,
        transformer: Union[Transformer, Type[ONNXTransformer]],
    ):
        orig_model = make_model(model_desc)
        copy_model = onnx.ModelProto()  # pylint: disable=no-member
        copy_model.CopyFrom(orig_model)

        if isinstance(transformer, Transformer):
            trans_model = transformer.transform(copy_model)
        elif issubclass(transformer, ONNXTransformer):
            trans_model = transformer(copy_model).transform()
        else:
            raise TypeError(repr(transformer))

        return orig_model, trans_model

    return inner


@pytest.fixture
def check_graph_node():
    def inner(model: onnx.ModelProto, op_types: List[str]):  # pylint: disable=no-member
        assert len(model.graph.node) == len(op_types)
        for node, op_type in zip(model.graph.node, op_types):
            assert node.op_type == op_type

    return inner


@pytest.fixture
def check_output_value():
    def inner(
        orig_model: onnx.ModelProto,  # pylint: disable=no-member
        trans_model: onnx.ModelProto,  # pylint: disable=no-member
        input_shapes: Iterable[Sequence[int]],
        data: Optional[List[np.ndarray]] = None,
    ):
        if data is None:
            rng = np.random.default_rng()
            data = [rng.standard_normal(shape, dtype=np.float32) for shape in input_shapes]

        actual = _run_onnx_model_flatten_output(orig_model, data)
        expected = _run_onnx_model_flatten_output(trans_model, data)

        for act, exp in zip(actual, expected):
            assert len(act) == len(exp)
            for a, b in zip(act, exp):
                assert a == pytest.approx(b, 2), f"{data}"

    return inner


# def check_attribute(actual, expected):
#     assert actual == expected


def _run_onnx_model(
    model: onnx.ModelProto, input_arrays: List[np.ndarray]  # pylint: disable=no-member
) -> List[np.ndarray]:
    sess = ort.InferenceSession(model.SerializeToString())
    input_names = [inp.name for inp in sess.get_inputs()]
    output_names = [out.name for out in sess.get_outputs()]
    feed_dict = dict(zip(input_names, input_arrays))
    outputs = sess.run(output_names, input_feed=feed_dict)

    return outputs


def _run_onnx_model_flatten_output(
    model: onnx.ModelProto, input_arrays: List[np.ndarray]  # pylint: disable=no-member
) -> List[List[float]]:
    outputs = _run_onnx_model(model, input_arrays)
    return [val.flatten().tolist() for val in outputs]



from furiosa.optimizer.frontend.onnx import optimize_model


def test_opset_version(make_test_model):
    model = make_test_model(opset_version=11)
    opset = 12
    model = optimize_model(model, opset_version=opset)

    assert model.opset_import[0].version, opset


def test_opset_version_1(make_test_model):
    model = make_test_model(opset_version=12)
    opset = 13
    model = optimize_model(model, opset_version=opset)

    assert model.opset_import[0].version, opset



from typing import Optional

import onnx
from onnx import version_converter

from furiosa.optimizer.frontend.onnx import __OPSET_VERSION__
from furiosa.optimizer.frontend.onnx.utils.check_model import check_model
from furiosa.optimizer.interfaces.transformer import Transformer


class CheckVersion(Transformer[onnx.ModelProto]):  # pylint: disable=no-member
    """Convert an ONNX model to ONNX opset 12 or 13"""

    opset_bound = [12, 13]

    def __init__(self, opset_version: Optional[int] = None) -> None:
        self.opset_version = opset_version or __OPSET_VERSION__

        if self.opset_version < self.opset_bound[0] or self.opset_version > self.opset_bound[1]:
            raise ValueError(
                f"Unsupported opset_version: {opset_version}. "
                "Choose {self.opset_bound[0]} <= opset_version <= {self.opset_bound[1]}."
            )

    def transform(self, model: onnx.ModelProto) -> onnx.ModelProto:  # pylint: disable=no-member
        version = int(model.opset_import[0].version)
        if version != self.opset_version:
            try:
                model = version_converter.convert_version(model, self.opset_version)
                check_model(model, check_runnable=False)
            except Exception as exc:
                raise NotImplementedError(
                    f"Can't convert the model (ONNX opset {version}) to ONNX opset "
                    "{__OPSET_VERSION__}"
                ) from exc
        return model



import numpy as np

from furiosa.optimizer.frontend.onnx.transformer.convert_prelu_to_relu import Pattern_1, Pattern_2
from furiosa.optimizer.frontend.onnx.transformer.utils import check_value_info


def test_case1(make_transformer_model, check_output_value, check_graph_node):
    input_shape = [1, 3, 4, 4]
    output_shape = [1, 3, 4, 4]
    opsetid = ("", 13)
    model_desc = {
        "input": {"data": (np.float32, input_shape)},
        "output": {"output": (np.float32, output_shape)},
        "initializer": {
            "slope": np.array([[[1]], [[-2]], [[3]]], dtype=np.float32),
        },
        "node": [
            ("PRelu", ["data", "slope"], ["output"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_1)

    check_output_value(orig_model, trans_model, [input_shape])
    check_graph_node(trans_model, op_types=["Relu", "Mul", "Mul", "Add"])
    check_value_info(trans_model)


def test_case2(make_transformer_model, check_output_value, check_graph_node):
    input_shape = [1, 3, 4, 4]
    slope_shape = [1, 3, 1, 1]
    output_shape = [1, 3, 4, 4]
    opsetid = ("", 13)
    model_desc = {
        "input": {
            "data": (np.float32, input_shape),
            "half_slope": (np.float32, slope_shape),
        },
        "output": {"output": (np.float32, output_shape)},
        "node": [
            ("Add", ["half_slope", "half_slope"], ["slope"]),
            ("PRelu", ["data", "slope"], ["output"]),
        ],
        "opsetid": [opsetid],
    }

    orig_model, trans_model = make_transformer_model(model_desc, Pattern_2)

    check_output_value(orig_model, trans_model, [input_shape, slope_shape])
    check_graph_node(trans_model, op_types=["Add", "Relu", "Sub", "Mul", "Mul", "Add"])
    check_value_info(trans_model)



import numpy as np
import onnx
import pytest

from furiosa.optimizer.frontend.onnx.transformer import utils


def test_shape_inference_no_dims(make_model):
    input_shape = [1, 4, 1, 1]
    output_shape = [4]
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "initializer": {},
        "node": [
            ("Squeeze", ["x"], ["x_1"]),
            ("Unsqueeze", ["x_1"], ["x_2"], {"axes": [0, 2, 3]}),
            ("Squeeze", ["x_2"], ["x_3"]),
            ("Unsqueeze", ["x_3"], ["x_4"], {"axes": [0, 2, 3]}),
            ("Squeeze", ["x_4"], ["y"]),
        ],
        "opsetid": [("", 12)],
    }
    model = make_model(model_desc)

    with pytest.raises(ValueError, match=r"shape of(\w)*"):
        utils.check_value_info(model)


def test_no_shape_inference(make_model):
    model_desc = {
        "input": {"x": (np.float32, [2, 2])},
        "output": {"z": (np.float32, [2, 2])},
        "node": [("Add", ["x", "x"], ["y"]), ("Add", ["x", "y"], ["z"])],
    }
    model = make_model(model_desc, infer_shape=False)

    with pytest.raises(ValueError, match=r"value_info of(\w)*"):
        utils.check_value_info(model)


def test_no_elem_type(make_model):
    def _make_y_value_info():
        value_info_y = onnx.ValueInfoProto()  # pylint: disable=no-member
        value_info_y.name = "y"
        type_proto_y = onnx.TypeProto()  # pylint: disable=no-member
        type_proto_tensor_y = onnx.TypeProto.Tensor()  # pylint: disable=no-member
        type_proto_y.tensor_type.CopyFrom(type_proto_tensor_y)
        tensor_shape_proto_y = onnx.TensorShapeProto()  # pylint: disable=no-member
        dim = onnx.TensorShapeProto.Dimension()  # pylint: disable=no-member
        dim.dim_value = 2
        tensor_shape_proto_y.dim.extend([dim, dim])
        type_proto_y.tensor_type.shape.CopyFrom(tensor_shape_proto_y)
        value_info_y.type.CopyFrom(type_proto_y)
        return value_info_y

    model_desc = {
        "input": {"x": (np.float32, [2, 2])},
        "output": {"z": (np.float32, [2, 2])},
        "node": [("Add", ["x", "x"], ["y"]), ("Add", ["x", "y"], ["z"])],
    }
    model = make_model(model_desc, infer_shape=False)
    model.graph.value_info.append(_make_y_value_info())

    with pytest.raises(ValueError, match=r"elem_type of(\w)*"):
        utils.check_value_info(model)



import numpy as np
import pytest

from furiosa.optimizer.frontend.onnx.utils.version_checker import CheckVersion


def test_case(make_model):
    input_shape = [8]
    output_shape = [8]
    opsetid = ("", 11)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "node": [("Relu", ["x"], ["y"])],
        "opsetid": [opsetid],
    }
    new_opset = 12
    orig_model = make_model(model_desc)
    # convert opset 11 to 12
    trans_model = CheckVersion(opset_version=new_opset).transform(orig_model)

    assert trans_model.opset_import[0].version == new_opset


def test_case1(make_model):
    input_shape = [8]
    output_shape = [8]
    opsetid = ("", 12)
    model_desc = {
        "input": {"x": (np.float32, input_shape)},
        "output": {"y": (np.float32, output_shape)},
        "node": [("Relu", ["x"], ["y"])],
        "opsetid": [opsetid],
    }
    new_opset = 13
    orig_model = make_model(model_desc)
    # convert opset 12 to 13
    trans_model = CheckVersion(opset_version=new_opset).transform(orig_model)

    assert trans_model.opset_import[0].version == new_opset


def test_case2():
    new_opset = 11

    # if target opset < 12, CheckVersion should raise ValueError
    with pytest.raises(ValueError):
        CheckVersion(new_opset)


def test_case3():
    new_opset = 14

    # if target opset > 13, CheckVersion should raise VlueError
    with pytest.raises(ValueError):
        CheckVersion(new_opset)
